#include <cstdio>       /* stderr          */
#include <cstdlib>      /* exit            */
#include <sys/socket.h> /* accept          */
#include <sys/types.h>  /* fd_set          */
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <vector>

#include "soundtypes.h"
#include "serverfun.h"
#include "mixerstream.hh"
#include "logfun.h"
#include "mixer.h"

class Mixer
{
    vector<MixerInput> inputs;
    vector<MixerOutput> outputs;
    vector<Waiter> waiters;
    int listfd;
    int humpid;
    int humfd;
public:
    Mixer(int port)
    {
        int pip[2];
        pipe(pip);
        humpid = fork();
        if(!humpid)
        {
            dup2(pip[1], 1);
            close(pip[0]); close(pip[1]);
            execlp("mixeridle", "mixeridle", NULL);
        }
        close(pip[1]);
        humfd = pip[0];
        
        listfd = MakeListenerSocket(port);
    }
    
    unsigned leastinputsize() const
    {
        unsigned least = 0x7FFFFFFFUL;
        vector<MixerInput>::const_iterator ii;
        for(ii=inputs.begin(); ii!=inputs.end(); ++ii)
        {
            fprintf(stderr, "Input %d has %u items to say\n", ii->getfd(), ii->size());
            if(ii->size() < least)
                least = ii->size();
        }
        return least;
    }
    
    void tick();
};

void Mixer::tick()
{
    fd_set ifds; FD_ZERO(&ifds);
    fd_set ofds; FD_ZERO(&ofds);
    int topfd = listfd;
    FD_SET(listfd, &ifds);
    
    bool panic = outputs.size() == 0;
    for(unsigned a=0; a<outputs.size(); ++a)
        if(outputs[a].panic())
        {
            /* panic tarkoittaa, ettei outputtiin mahdu lis tavaraa.
             * tllin pit pit inputit holdissa ja odotella kunnes
             * outputti on synyt sen mit sille on sytetty.
             */
            panic = true;
        }
    if(!panic || !outputs.size())
        for(unsigned a=0; a<inputs.size(); ++a)
        {
            if(inputs[a].Evaluatable())
            {
                int fd = inputs[a].getfd();
                if(fd > topfd)topfd = fd;
                FD_SET(fd, &ifds);
            }
            else
            {
                fprintf(stderr, "Input %d doesn't want feed\n",
                    inputs[a].getfd());
            }
        }
    
    for(unsigned a=0; a<outputs.size(); ++a)
    {
        int fd = outputs[a].getfd();
        if(fd > topfd)topfd = fd;
        if(outputs[a].pending()
        || !inputs.size())
        {
            FD_SET(fd, &ofds);
        }
    }
    for(unsigned a=0; a<waiters.size(); ++a)
    {
        int fd = waiters[a].getfd();
        if(fd > topfd)topfd = fd;
        FD_SET(fd, &ifds);
    }
    
    select(topfd+1, &ifds, &ofds, NULL, NULL);
    
    if(FD_ISSET(listfd, &ifds))
    {
        struct sockaddr_in They;
        unsigned AddrLen = sizeof They;
        
        int p = accept(listfd, (struct sockaddr *)&They, &AddrLen);

        int bufsize = 8192;
        setsockopt(p, SOL_SOCKET, SO_RCVBUF, &bufsize, sizeof bufsize);
        setsockopt(p, SOL_SOCKET, SO_SNDBUF, &bufsize, sizeof bufsize);        
        
        //DumpTime();
        //fprintf(stderr, "Got connection (fd %d)\n", p);
        
        waiters.push_back(Waiter(p, They, AddrLen));
    }
    
    for(unsigned a=0; a<inputs.size(); ++a)
        if(FD_ISSET(inputs[a].getfd(), &ifds))
            inputs[a].Evaluate();
        else
        {
            fprintf(stderr, "Input %d is idle\n", inputs[a].getfd());
        }

    for(unsigned a=0; a<waiters.size(); ++a)
        if(FD_ISSET(waiters[a].getfd(), &ifds))
            waiters[a].Evaluate();

    for(unsigned a=0; a<outputs.size(); ++a)
        if(FD_ISSET(outputs[a].getfd(), &ofds))
            outputs[a].Evaluate();
    
    vector<Waiter>::iterator wi;
    for(wi = waiters.begin(); wi != waiters.end(); )
        if(wi->isinput()) { inputs.push_back(*wi); waiters.erase(wi); }
        else if(wi->isoutput()) { outputs.push_back(*wi); waiters.erase(wi); }
        else ++wi;
        
    bool problems = false;

    vector<MixerInput>::iterator ii;
    for(ii=inputs.begin(); ii!=inputs.end(); ) if(ii->problems()) { /*fprintf(stderr, "In killed\n");*/ problems = true; inputs.erase(ii); } else ++ii;
    
    vector<MixerOutput>::iterator oi;
    for(oi=outputs.begin(); oi!=outputs.end(); ) if(oi->problems()) { /*fprintf(stderr, "Out killed\n");*/ problems = true; outputs.erase(oi); } else ++oi;
    
    if(problems)
        waitpid(-1, NULL, WNOHANG);
    
    if(inputs.size())
    {
        unsigned least = leastinputsize();
        /* If something to send */
        if(least)
        {
            unsigned Bytes = least * sizeof(stent);
            stent *Buf = (stent *) alloca(Bytes);
            memset(Buf, 0, Bytes);
            
            for(ii=inputs.begin(); ii!=inputs.end(); ++ii)
            {
                unsigned a;
                const stent *source = ii->ptr();
                for(a=0; a<least; ++a)
                {
                    Buf[a].l += source[a].l;
                    Buf[a].r += source[a].r;
                }
                ii->Eat(least);
            }
            
            for(oi=outputs.begin(); oi!=outputs.end(); ++oi)
            {
                fprintf(stderr, "Sending %u items to output %d\n", least, oi->getfd());
                oi->Send(Buf, least);
            }
        }
    }
    else if(outputs.size() && !panic)
    {
        /* No inputs! Send hum. */
        char Buf[4096];
        int c = read(humfd, Buf, sizeof Buf);
        fprintf(stderr, "Humming %d bytes\n", c);
        if(c > 0)
        {
            for(oi=outputs.begin(); oi!=outputs.end(); ++oi)
                if(!oi->panic())
                    oi->Send((stent *)&Buf[0], c/4);
        }
    }
}

static void Killer(int)
{
    exit(-1);
}

#include <argh.hh>

#define VERSIONSTR \
    "mixer v"VERSION" - Copyright (C) 1992,2002 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "This program is under GPL. streamixer-"VERSION".{rar,zip,tar.{gz,bz2}}\n" \
    "are available at the homepage of the author.\n"

int main(int argc, const char *const *argv)
{
    ParamHandler Argh;
    Argh.AddLong("help",    'h').SetBool().SetDesc("This text");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Version info");
    Argh.StartParse(argc, argv);
    for(int c; (c=Argh.GetParam()) != -1; )
        switch(c)
        {
            case 'V':
                printf(VERSIONSTR);
                return 0;
            case 'h':
                printf(
                    "%s\n"
                    "Usage: %s [<options> ...]\n"
                    "\n"
                    "Provides a server for audio stream mixing.\n"
                    "Meant to be a replacement for esd, with much more features\n"
                    "and much less cpu usage. See mixwrite -h, mixmon -h\n"
                    "\n"
                    "Options for mixer:\n", VERSIONSTR, argv[0]);
                Argh.ListOptions();
                return 0;
        }
    if(!Argh.ok())return -1;
    
    signal(SIGPIPE, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGTERM, Killer);
    
    Mixer mixer(MIXER_PORT);
    
    for(;;)
    {
        mixer.tick();
    }
    
    return 0;
}
