#include <cstdio>
#include <cstring>
#include <string>
#include <list>

#include "soundtypes.h"

using namespace std;

static stent operator* (const stent &a, double b)
{
	int l((int)(a.l * b)); if(l < -32768) l = -32768; if(l > 32767) l = 32767;
	int r((int)(a.r * b)); if(r < -32768) r = -32768; if(r > 32767) r = 32767;
	stent tmp = {l,r};
	return tmp;
}
static void operator+= (stent &a, const stent &b)
{
	int l(a.l + b.l); if(l < -32768) l = -32768; if(l > 32767) l = 32767;
	int r(a.r + b.r); if(r < -32768) r = -32768; if(r > 32767) r = 32767;
	
	a.l = l;
	a.r = r;
}

class Job
{
	string command;
	double starttime;
	double volume;
	
	bool started;
	bool ended;
	FILE *fp;
	
	void Start()
	{
		if(started)return;
		fprintf(stderr, "Launching %s...\n", command.c_str());
		fp = popen(command.c_str(), "r");
		started = true;
		ended = false;
	}
	void End()
	{
		if(!started)return;
		pclose(fp);
		started = false;
	}
	
public:	
	Job(const string &cmd, double tim, double power=1)
	: command(cmd), starttime(tim), volume(power),
	  started(false), ended(false), fp(NULL)
	{
	}
	~Job()
	{
		End();
	}
	
	inline bool iseof() const { return started && ended; }
	double getstarttime() const { return starttime; }
	double getvolume() const { return volume; }

	unsigned Read(stent *dest, unsigned count)
	{
		if(ended) return 0;
		if(!started)Start();
		
		if(feof(fp) || ferror(fp)) { ended = true; return 0; }

		return fread((void *)dest, sizeof(*dest), count, fp);
	}
};

static double lomitus;

class Mixer
{
	typedef list<Job> joblist;
	joblist jobs;
	unsigned rate;

	void Write(stent *src, unsigned count)
	{
		fwrite((void *)src, sizeof(*src), count, stdout);
	}
public:
	Mixer(unsigned freq) : rate(freq)
	{
	}
	void AddJob(const Job &job)
	{
		jobs.push_back(job);
	}
	void AddJob(const string &cmd, double at)
	{
		AddJob(Job(cmd, at));
	}
	void Mix()
	{
		double clock = 0;
		for(;;)
		{
			double nextstart = clock + lomitus;
			joblist::iterator i;
			for(i=jobs.begin(); i!=jobs.end(); )
			{
				if(i->iseof()) 
				{
					joblist::iterator j = i; ++j;
					jobs.erase(i); i = j; continue;
				} 
				
				if(i->getstarttime() > clock
				&& i->getstarttime() < nextstart)
				{
					nextstart = i->getstarttime();
				}
				++i;
			}
			
			if(!jobs.size()) break;
			
			double runlength = nextstart - clock;
			
			unsigned samples = (unsigned)(rate * runlength + 0.5);
			if(!samples)
			{
				clock += runlength;
				continue;
			}
			
			stent *Buf = new stent[samples];
			stent *SubBuf = new stent[samples];
			memset((void *)Buf, 0, samples*sizeof(*Buf));
			
			unsigned maxread=0;
			for(i=jobs.begin(); i!=jobs.end(); ++i)
			{
				if(i->getstarttime() <= clock)
				{
					double vol = i->getvolume();
					
					unsigned n = i->Read(SubBuf, samples);
					for(unsigned a=0; a<n; ++a)
						Buf[a] += SubBuf[a] * vol;
					if(n > maxread) maxread = n;
				}
			}
			if(!maxread) maxread = samples;
			fprintf(stderr, "Writing %u samples at %g\n", maxread, clock);
			Write(Buf, maxread);
			
			delete[] Buf;
			delete[] SubBuf;

			clock += maxread / (double)rate;
		}
		fprintf(stderr, "Mixed %g seconds\n", clock);
	}
};

#include <argh.hh>

#define VERSIONSTR \
    "mixerscript v"VERSION" - Copyright (C) 1992,2003 Bisqwit (http://bisqwit.iki.fi/)\n" \
    "This program is under GPL. streamixer-"VERSION".tar.{gz,bz2}\n" \
    "are available at the homepage of the author.\n"

int main(int argc, const char *const *argv)
{
	ParamHandler Argh;
	
    Argh.AddLong("help",     'h').SetBool().SetDesc("Help");
    Argh.AddLong("version",  'V').SetBool().SetDesc("Version information");
    Argh.AddLong("volume",   'v').SetFloat(-1,100).SetDesc("Set mixing volume of the next element (default: 1)", "<float>");
    Argh.AddLong("at",       't').SetFloat(0,1e99).SetDesc("Set time offset of the next element (default: 0)", "<float>");
    Argh.AddLong("blocksize",'b').SetFloat(0.1,3600).SetDesc("Set maxblocksize (default: 0.2) - experts only", "<float>");
    Argh.AddLong("rate",     'r').SetInt(1,999999).SetDesc("Set global sampling rate in hertz (defalt: 44100)", "<int>");
    
    bool Heelp = false;
    unsigned rate = 44100;
    double volume = 1.0;
    double pos = 0.0;
    
    lomitus = 0.2;
    
    list<Job> jobs;
    
    Argh.StartParse(argc, argv);
    for(int c; (c=Argh.GetParam()) != -1; )
        switch(c)
        {
            case 'h': Heelp = Argh.GetBool(); break;
            case 'V':
                printf("%s", VERSIONSTR);
                return 0;
            case 'v': volume = Argh.GetFloat(); break;
            case 't': pos    = Argh.GetFloat(); break;
            case 'r': rate   = Argh.GetInt(); break;
            case 'b': lomitus= Argh.GetFloat(); break;
            default:
            {
            	fprintf(stderr, "Adding job: '%s' (pos %g s, volume %g%%)\n",
            		Argh.GetString().c_str(), pos, volume*100.0);
            	
            	Job tmp(Argh.GetString(), pos, volume);
            	jobs.push_back(tmp);
                break;
            }
        }
     
    if(!Argh.ok())return -1;
    if(Heelp || !jobs.size())
    {
        fprintf(stderr,
            "%s\n"
            "Usage: %s {<option> | <command>}\n"
            "\n"
            "Writes stdout.\n"
            "\n"
            "Note: All streams must be 16-bit stereo.\n"
            "      Pipe the stream with resample program if this isn't the case...\n"
            "Specify streams as commands which output the data.\n"
            "\n"
            "Example: mixerscript -t0   'sox tmp.wav -t raw -' \\\n"
            "                     -t0.2 'mpg123 -s test.mp3' > test.raw \n"
            "\nOptions:\n", VERSIONSTR, argv[0]);
        
        Argh.ListOptions();
        return 0;
    }
    
	Mixer mix(rate);
	
	list<Job>::const_iterator i;
	for(i=jobs.begin(); i!=jobs.end(); ++i)
    	mix.AddJob(*i);
    
    fprintf(stderr, "Mixing %u stream(s)...\n", jobs.size());
	
	mix.Mix();
	
	fprintf(stderr, "Done\n");
	
	return 0;
}
