#include <cerrno>
#include <cstdio>
#include <cctype>
#include <cstring>     /* for memset */
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h> /* for inet_ntoa */
#include <unistd.h>
#include <fcntl.h>
#include <string>

using namespace std;

#include "mixer.h"

#define VERSIONSTR \
    "%s v"VERSION" - Copyright (C) 1992,2002 Bisqwit (http://iki.fi/bisqwit/)\n" \
    "This program is under GPL. streamixer-"VERSION".{rar,zip,tar.{gz,bz2}}\n" \
    "are available at the homepage of the author.\n", Argh.ProgName().c_str()

string Host = "localhost";
string StreamForm;

static string GetError(int fd)
{
    char c = 0;
    read(fd, &c, 1);
    if(c != '!') /* Get error message if any */
    {
        string err;
        do {
            err += c;
        } while(read(fd, &c, 1) == 1);
        return err;
    }
    return "";
}

#include <argh.hh>

static int blur = 1;
static string filter;
static const string GetFilter()
{
    char Buf[64] = "";
    if(blur < 1)
        sprintf(Buf, "s%d", 1-blur);
    else if(blur > 1)
        sprintf(Buf, "b%d", blur-1);
    string p = Buf;
    if(filter.size())p += "f" + filter;
    if(!p.size())return "";
    return "-"+p;
}

int main(int argc, const char *const *argv)
{
	ParamHandler Argh;
	
	/* No descs. Intentional. */
	Argh.AddLong("help",      'h').SetBool();
	Argh.AddLong("version",   'V').SetBool();
	Argh.AddLong("host",      'H').SetString().SetDesc("Specify host where mixer is running", "<host>");
	Argh.AddLong("filter",    'f').SetString();
	Argh.AddLong("format",    'F').SetString().SetDesc("Stream format string, see above", "<format>");
	Argh.AddLong("blur",      'b').SetInt(1,999);
	Argh.AddLong("sharpen",   's').SetInt(1,999);

    bool Heelp=false;
    Argh.StartParse(argc, argv);
    for(int c; (c=Argh.GetParam()) != -1; )
    	switch(c)
    	{
    		case 'V':
		        printf(VERSIONSTR);
		        return 0;
		    case 'h': Heelp = Argh.GetBool(); break;
		    case 'f': filter = Argh.GetString(); break;
		    case 'H': Host = Argh.GetString(); break;
		    case 'F': StreamForm = Argh.GetString(); break;
    		case 'b': blur = 1 + Argh.GetInt(); break;
    		case 's': blur = 1 - Argh.GetInt(); break;
    		default:
    		    break;
    	}

	char type = (Argh.ProgName() != "mixmon")
                   ? 'S'  /* stream */
                   : 'M'; /* monitor */
    
    int fd;
    
    if(Heelp) /* just want the help text */
    {
        int pip[2];
        pipe(pip);
        int pid = fork();
        if(!pid)
        {
            close(pip[0]);
            dup2(pip[1], 1);
            dup2(pip[1], 2);
            close(pip[1]);
            execlp("resample", "resample", "--help", NULL);
            fprintf(stderr, VERSIONSTR);
            fprintf(stderr,
            	"\n"
            	"resample not found.\n"
            	"It is required. It is normally included in the\n"
            	"same package where this program came from.\n"
            );
            _exit(-1);
        }
        close(pip[1]);
        fd = pip[0];
    }
    else
    {
        struct sockaddr_in sin;
        sin.sin_port = htons(MIXER_PORT);
        struct hostent *hent = gethostbyname(Host.c_str());
        if(!hent)
        {
            fprintf(stderr, "Can't resolve hostname \"%s\"\n", Host.c_str());
            return -1;
        }
        memcpy(&sin.sin_addr, hent->h_addr_list[0], sizeof(struct in_addr));
        fd = socket(sin.sin_family = AF_INET, SOCK_STREAM, 0);
        if(connect(fd, (struct sockaddr *)&sin, sizeof(struct sockaddr)) < 0)
        {
            close(fd);
            fprintf(stderr, "Unable to connect to \"%s\"", Host.c_str());
            perror("");
            return errno;
        }

	   	int bufsize = 32768;
	    setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &bufsize, sizeof bufsize);
    	setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &bufsize, sizeof bufsize);
        
        static const char lf = '\n';
        write(fd, &type, 1);
        string FilterParm = GetFilter();
        write(fd, StreamForm.data(), StreamForm.size()); write(fd, &lf, 1);
        write(fd, FilterParm.data(), FilterParm.size()); write(fd, &lf, 1);
    }
    
    string error = GetError(fd);
    if(error.size())
    {
        unsigned spec = error.find("spec:");
        unsigned opts = error.find("Options:");
        if(spec == error.npos
        || opts == error.npos)
        {
            fprintf(stderr, "%s", error.c_str());
            return -1;
        }
        string specs = error.substr(spec+6,opts-spec-8);
        string ops = error.substr(opts+9, error.size()-opts-10);
        
        fprintf(stderr,
            "Usage: %s [<options> ...]\n"
            "\n"
            "%s.\n"
            "\n",
            Argh.ProgName().c_str(),
            (Argh.ProgName()=="mixmon" ? "Writes stdout" : "Reads stdin"));
        fprintf(stderr,
            "Stream format strings (examples):\n"
            "%s\n"
            "\n"
            "Options:\n", specs.c_str());
        Argh.ListOptions();
        fprintf(stderr, "%s\n", ops.c_str());

        return -1;
    }
    
    int src  = type=='S' ? 0 : fd;
    int dest = type=='M' ? 1 : fd;
    
    for(;;)
    {
        char Buf[4096];
        int c = read(src, Buf, sizeof Buf);
        if(c <= 0)break;
        
        /* We get SIGPIPE when this fails, so no worth checking */
        write(dest, Buf, c);
    }
        
    return 0;
}
