#include <csignal>
#include <cstring>
#include <cstdio>
#include <cctype>
#include <algorithm>
#include <unistd.h>

using namespace std;

#include "serverfun.h"
#include "resample.hh"
#include "filters.h"

static unsigned Filter = 1;  /* linear */
static double blur = 1;      /* <1 = sharp, >1 = blurry */

/* This the function which started the program development.
 * The core of this program. The rest is just waveform
 * conversions and parameter handling...
 */
static void Interpolate(Stream &input, Stream &output)
{
    if(input.rate() == output.rate())
    {
        // No change in rate, just copy
        stent tab[8192];
        for(;;)
        {
            unsigned oco = input.Read(tab, 8192);
            if(!oco)return;
            output.Write(tab, 8192);
        }
    }
    
    double factor = (double)input.rate() / output.rate();
    
    double scale_factor = blur * max(factor, 1.0);
    double support = max(scale_factor * Filters[Filter].support, 0.5);
    if(support <= 0.5)
    {
        /* reduce to point sampling (??) */
        support = 0.5 + 1E-12;
        scale_factor = 1.0;
    }
    
    double *weight = (double *)alloca((sizeof(double)) * ((int)support*2+3));
    double (*const filterfun)(double) = Filters[Filter].function;
    
    double center = 0.0;
    const unsigned outbufsize = output.rate() / 8;
    
    stent *oent = (stent *)alloca(outbufsize * sizeof(stent));
    
    /* This resampling algorithm is based on ideas from ImageMagick. */
    for(;;)
    {
        for(unsigned o=0; o<outbufsize; ++o, center += factor)
        {
            int start = (int)(0.5 + center - support);
            int end   = (int)(0.5 + center + support);
            
            register double density = 0;
            int n = 0;
            for(int i = start; i < end; ++i)
                density += (weight[n++] = filterfun((i-center+0.5) / scale_factor) / scale_factor);
            
            if(density != 0.0)
                for(register int i=0; i<n; ++i)
                    weight[i] /= density; /* normalize */
            
            register double l=0.0;
            register double r=0.0;
            for(register int i=0, p=start; i<n; ++i)
            {
                const stent &srcent = input[p++];
                l += weight[i] * srcent.l;
                r += weight[i] * srcent.r;
            }
            oent[o].l = l<-32768 ? -32768 : l>32767 ? 32767 : (int)(l+0.5);
            oent[o].r = r<-32768 ? -32768 : r>32767 ? 32767 : (int)(r+0.5);
        }

        output.Write(oent, outbufsize);
        output.Flush();
        
        register int shiftval = (int)center;
        /* Prevent overflowing in 17 hours */
        input.shift(shiftval);
        center -= shiftval;
    }
}

Stream::Stream(const char *settings) : bit16(true), stereo(true),
                                       herz(0), basepos(-inbufsize)
{
    while(*settings)
    {
        if(*settings=='m')stereo = false;
        else if(*settings=='s')stereo = true;
        else if(*settings=='b')bit16 = false;
        else if(*settings=='w')bit16 = true;
        else if(*settings=='r')herz = 0;
        else if(*settings>='0' && *settings<='9')
            herz = herz * 10 + (*settings - '0');
        else if(*settings!='-')throw *settings;
        ++settings;
    }
    if(!herz)herz = 44100;
}

const stent &Stream::operator[] (int index)
{
    while(index >= basepos + inbufsize)
    {
        basepos += leapstep;
        memmove(&inent[0], &inent[leapstep], min(leapstep,newsize) * sizeof(inent[0]));
        int rval = Read(&inent[leapstep], newsize);
        if(!rval)
        {
            /* EOF */
            exit(0);
        }
        if(rval < newsize)
            memset(&inent[leapstep+rval], 0, (newsize-rval)*sizeof(inent[0]));
    }
    if(index < basepos)return inent[0];
    return inent[index - basepos];
}

unsigned Stream::Read(stent *dest, unsigned count) const
{
    if(bit16)
        if(stereo)
            return SafeRead(dest, count*4) / 4;
        else
            return read16(dest, count);
    else
        if(stereo)
            return readste(dest, count);
        else
            return read8(dest, count);
}
void Stream::Write(const stent *dest, unsigned count) const
{
    if(bit16)
        if(stereo)
            SafeWrite(dest, count*4);
        else
            write16(dest, count);
    else
        if(stereo)
            writeste(dest, count);
        else
            write8(dest, count);
}

void Stream::Flush() const
{
    fdatasync(1);
}

/* Read 16-bit mono entries */
unsigned Stream::read16(stent *dest, unsigned count)
{
    unsigned bytes = count * sizeof(int16);
    int16 *tab = (int16 *)alloca(bytes);
    count = SafeRead(tab, bytes) / 2;
    for(unsigned a=0; a<count; ++a)
    {
        dest[a].r = dest[a].l = tab[a];
    }
    return count;
}
/* Read 8-bit stereo entries */
unsigned Stream::readste(stent *dest, unsigned count)
{
    unsigned bytes = count * 2 * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    count = SafeRead(tab, bytes) / 2;
    for(unsigned a=0, b=0; b<count; ++b)
    {
        dest[b].r = tab[a++]*256 - 32768;
        dest[b].l = tab[a++]*256 - 32768;
    }
    return count;
}
/* Read 8-bit mono entries */
unsigned Stream::read8(stent *dest, unsigned count)
{
    unsigned bytes = count * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    count = SafeRead(tab, bytes);
    for(unsigned a=0; a<count; ++a)
    {
        dest[a].r = dest[a].l = tab[a]*256 - 32768;
    }
    return count;
}

void Stream::write16(const stent *src, unsigned count)
{
    unsigned bytes = count * sizeof(int16);
    int16 *tab = (int16 *)alloca(bytes);
    for(unsigned a=0; a<count; ++a)
    {
        /* 16-bit stereo to mono mixing */
        tab[a] = (src[a].r + src[a].l) / 2;
    }
    SafeWrite(tab, bytes);
}
void Stream::writeste(const stent *src, unsigned count)
{
    unsigned bytes = count * 2 * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    for(unsigned a=0, b=0; b<count; ++b)
    {
        tab[a++] = (src[b].r + 32768) / 256;
        tab[a++] = (src[b].l + 32768) / 256;
    }
    SafeWrite(tab, bytes);
    
}
void Stream::write8(const stent *src, unsigned count)
{
    unsigned bytes = count * sizeof(uint8);
    uint8 *tab = (uint8 *)alloca(bytes);
    for(unsigned a=0; a<count; ++a)
    {
        /* 8-bit stereo to mono mixing */
        tab[a] = ((src[a].r + src[a].l) / 2 + 32768) / 256;
    }
    SafeWrite(tab, bytes);
}

#define VERSIONSTR \
    "resample v"VERSION" - Copyright (C) 1992,2003 Bisqwit (http://bisqwit.iki.fi/)\n" \
    "Portions copyright (C) 2001 ImageMagick Studio.\n" \
    "This program is under GPL. streamixer-"VERSION".tar.{gz,bz2}\n" \
    "are available at the homepage of the author.\n"

static void Killer(int)
{
    exit(-1);
}

#include <argh.hh>

int main(int argc, const char *const *argv)
{
    signal(SIGPIPE, Killer);
    
    ParamHandler Argh;
    
    Argh.AddLong("help",     'h').SetBool().SetDesc("Help");
    Argh.AddLong("version",  'V').SetBool().SetDesc("Version information");
    Argh.AddLong("blur",     'b').SetInt(1,999).SetDesc("blur: 0=normal, >0=blur\n(note: blurring requires cpu power.)", "<num>");
    Argh.AddLong("sharpen",  's').SetInt(1,999).SetDesc("sharpen: 0=normal, >0=sharper\n(note: blurring and sharpening are exclusive.)", "<num>");
    Argh.AddBool('H');    
    string FiltText = "Select filter of the following:\n";
    for(unsigned p=0, a=0; a<FilterCount; ++a)
    {
        string L = Filters[a].name;
        if(a==Filter)L += " (default)";
        if(p + L.size() > 30) { p = 0; FiltText += "\n"; } else if(a > 0)FiltText += ' ';
        if(!p)FiltText += "   ";
        FiltText += L;
        p += L.size();
        if(a+1 < FilterCount)FiltText += ',';
    }
    Argh.AddLong("filter",   'f').SetString().SetDesc(FiltText, "<filter>");
    
    bool Heelp = false;
    bool stderr_helper = false;
    
    Stream *input = NULL, *output = NULL;
    if(argc > 2)
    {
        try
        {
            input = new Stream(argv[1]);
            output = new Stream(argv[2]);
        }
        catch(char c)
        {
            fprintf(stderr, "Invalid option: '%c' (-switches must be specified after stream parameters)\n", c);
            return -1;
        }
    }
    
    Argh.StartParse(argc, argv, argc>2 ? 3 : 1);
    for(int c; (c=Argh.GetParam()) != -1; )
        switch(c)
        {
            case 'H': stderr_helper = true; break;
            case 'h': Heelp = Argh.GetBool(); break;
            case 'V':
                printf("%s", VERSIONSTR);
                return 0;
            case 'b': blur = 1 + Argh.GetInt(); break;
            case 's': blur = 1 - Argh.GetInt(); break;
            case 'f':
            {
                string s = Argh.GetString();
                unsigned a;
                for(a=0; a<FilterCount; ++a)
                    if(Filters[a].name == s)
                    {
                        Filter = a;
                        break;
                    }
                if(a < FilterCount)break;
                fprintf(stderr, "'%s' is not a supported filter.\n", s.c_str());
                return -1;
            }
            default:
                break;
        }
    if(!Argh.ok())return -1;
    if(Heelp || argc <= 2)
    {
        fprintf(stderr,
            "%s\n"
            "Usage: %s <inputspec> <outputspec> [<options> ...]\n"
            "\n"
            "Reads stdin, writes stdout.\n"
            "\n"
            "input- and outputspec:\n"
            "   b=8-bit, m=mono, rxxx=xxx Hz, examples:\n"
          "       \"\"      = 16-bit 44100 Hz stereo\n"
            "       m       = 16-bit 44100 Hz mono\n"
            "       b       = 8-bit 44100 Hz stereo\n"
            "       mb      = 8-bit 44100 Hz mono\n"
            "       mr22050 = 16-bit 22050 Hz mono\n"
            "       br22050 = 8-bit 22050 Hz stereo\n"
            "   etc. '-' is optional.\n"
            "\nOptions:\n", VERSIONSTR, argv[0]);
        
        Argh.ListOptions();
        return 0;
    }
    
    if(stderr_helper)
    {
        static const char c = '!';
        write(2, &c, 1);
    }
    Interpolate(*input, *output);
    return 0;
}
